/*
 LangloPackets.h - Langlo LoRa network packet structures
 Version 1.0.0	06/11/2019
 Version 1.0.1	23/11/2019	Reduce max packet size to 64 (payload <= 51 per LoRaWAN)
 Version 1.0.2	04/02/2020	Add voltageTypePacket
 Version 1.1.0  28/03/2020  Replace all float types with int/uint
 
 Created by Peter Harrison, November 2019
*/
 
#ifndef LangloPackets_h
#define LangloPackets_h

// Langlo LoRa packet structure

const int LGO_PacketSize = 64;
const int LGO_HeaderSize = 16;
const int LGO_PayloadSize = 48;  // LGO_PacketSize - LGO_HeaderSize

union dataPacket {
  uint8_t packetByte[LGO_PacketSize];
  struct packetContent {
    uint32_t destinationMAC;        // 4 bytes
    uint32_t sourceMAC;             // 4 bytes
    uint32_t checksum;              // 4 bytes
    uint16_t sequenceNumber;        // 2 bytes
    uint8_t type;                   // 1 byte
    uint8_t byteCount;              // 1 byte
    uint8_t payload[LGO_PayloadSize];  // max 48 bytes
  } content;
};

const int powerTypePacket         = 0x00;
const int voltageTypePacket		  = 0x01;
const int tankTypePacket          = 0x11;
const int pumpTypePacket          = 0x12;
const int weatherTypePacket       = 0x20;
const int atmosphereTypePacket    = 0x21;
const int rainfallTypePacket      = 0x22;
const int windTypePacket          = 0x23;
const int sprinklerTypePacket     = 0x30;

const int powerPayloadBytes       = 0x0C;
const int voltagePayloadBytes	  = 0x02;
const int tankPayloadBytes        = 0x01;
const int pumpPayloadBytes        = 0x01;
const int weatherPayloadBytes     = 0x0B;
const int atmospherePayloadBytes  = 0x06;
const int rainfallPayloadBytes    = 0x02;
const int windPayloadBytes        = 0x03;
const int sprinklerPayloadBytes   = 0x30;
const int defaultPayloadBytes     = 0x30;

// Power Type

union powerPayload {
  uint8_t payloadByte[12];
  struct payloadContent {
  	uint16_t voltage;
  	uint16_t current;
  } battery, panel, load;
};

// Voltage Type

union voltagePayload {
  uint8_t payloadByte[2];
  struct payloadContent {
  	uint16_t voltage;
  } source;
};

// Tank Type

union tankPayload {
  uint8_t payloadByte[1];
  struct payloadContent {
    uint8_t level;				// struct will allow flexibility to add more parameters
  } tank;
};

// Pump Type

union pumpPayload {
  uint8_t payloadByte[1];
  struct payloadContent {
    uint8_t on;					// struct will allow flexibility to add more parameters
  } power;
};

// Weather Type

union weatherPayload {
  uint8_t payloadByte[11];
  struct payloadContent {
    uint16_t temperature;
    uint16_t pressure;
    uint16_t humidity;
    uint16_t rainfall;
    uint16_t windDirection;
    uint8_t windSpeed;
  } recorded;
};

// Atmosphere Type

union atmospherePayload {
  uint8_t payloadByte[6];
  struct payloadContent {
    uint16_t temperature;
    uint16_t pressure;
    uint16_t humidity;
  } recorded;
};

// Rainfall Type

union rainfallPayload {
  uint8_t payloadByte[2];
  struct payloadContent {
    uint16_t rainfall;
  } recorded;
};

// Wind Type

union windPayload {
  uint8_t payloadByte[3];
  struct payloadContent {
    uint16_t windDirection;
    uint8_t windSpeed;
  } recorded;
};

// Controller Type

union controllerPayload {
	uint8_t payloadByte[LGO_PayloadSize];			// No structure defined at this time
};

// Generic Type

union genericPayload {
  uint8_t payload[LGO_PayloadSize];
  powerPayload power;
  voltagePayload voltage;
  tankPayload tank;
  pumpPayload pump;
  weatherPayload weather;
  atmospherePayload atmosphere;
  rainfallPayload rainfall;
  windPayload wind;
  controllerPayload controller;
};

#endif
