/* Weather Station Wind Direction & Speed Node
 *
 * Function:
 * Read wind sensors and forward data to LoRa Gateway
 * 
 * Description:
 * 1. Based on Heltec CubeCell pingpong code
 * 2. Uses Langlo Packet structures;
 * 3. Reads wind direction and speed sensors
 * 4. Communicates data to [Heltec WiFi LoRa 32] LoRa Gateway;
 *
 * Digital Concepts
 * www.digitalconcepts.net.au
 *
 */

#include "LoRaWan_APP.h"
#include "Arduino.h"

#include "CRC32.h"            // Checksum methods
#include "LangloLoRa.h"       // LoRa configuration parameters
#include "LangloPackets.h"    // Packet structures
#include "nodeW1config.h"     // Node-specific configuration parameters

/*
 * set LoraWan_RGB to 1,the RGB active in loraWan
 * RGB red means sending;
 * RGB green means received done;
 */
#ifndef LoraWan_RGB
#define LoraWan_RGB 0
#endif

/* 
 *  The following parameters need to be synchronised with those specified in LangloLoRa.h, the latter being used
 *  for all processors other than the CubeCell
 *  
 *  #define BAND 917E6
 *  #define SignalBandwidth 125E3
 *  #define SpreadingFactor 7
 *  #define CodingRateDenominator 5
 *  #define PreambleLength 8
 *  
 *  At this stage, I dont know where the CubeCell software sets its SyncWord, or whether it's just set at 0x12
 */
#define RF_FREQUENCY                                917000000 // Hz
#define LORA_BANDWIDTH                              0         // [0: 125 kHz,
                                                              //  1: 250 kHz,
                                                              //  2: 500 kHz,
                                                              //  3: Reserved]
#define LORA_SPREADING_FACTOR                       7         // [SF7..SF12]
#define LORA_CODINGRATE                             1         // [1: 4/5,
                                                              //  2: 4/6,
                                                              //  3: 4/7,
                                                              //  4: 4/8]
#define LORA_PREAMBLE_LENGTH                        8         // Same for Tx and Rx

#define TX_OUTPUT_POWER                             14        // dBm
#define LORA_SYMBOL_TIMEOUT                         0         // Symbols
#define LORA_FIX_LENGTH_PAYLOAD_ON                  false
#define LORA_IQ_INVERSION_ON                        false


#define RX_TIMEOUT_VALUE                            1000
#define BUFFER_SIZE                                 64 // Define the payload size here

// char txpacket[BUFFER_SIZE];
// char rxpacket[BUFFER_SIZE];

uint8_t txpacket[BUFFER_SIZE];
uint8_t rxpacket[BUFFER_SIZE];

static RadioEvents_t RadioEvents;
void OnTxDone( void );
void OnTxTimeout( void );
void OnRxDone( uint8_t *payload, uint16_t size, int16_t rssi, int8_t snr );

typedef enum
{
    LOWPOWER,
    RX,
    TX
}States_t;

int16_t txnumber;
States_t state;
bool sleepmode = false;
int16_t RSSI,rxSize;

// Packet structures
dataPacket outgoingPacket = {0};
windPayload packetPayload = {0};

// Maintain a running average of <valueDimension> readings
uint8_t humidity = 0;
uint16_t pressure = 0;
int temperature = 0;

uint16_t windDirection = 0;
uint8_t windSpeed = 0;

const float LSB = 0.02761647; // per manual calibration (when using int in 1/10mm)
uint8_t inputRainfall = 0;
int adc0 = 0;

const int reportInterval = 60000;   // Report values every 60000 milliseconds (60 seconds)

int messageCounter = 0;

void setup() {
  BoardInitMcu( );
  Serial.begin(115200);
  Serial.print("[setup] ");
  Serial.print(myName);
  Serial.println(" Wind/LoRa Sensor");
  Serial.println();

    txnumber=0;
    RSSI=0;

    RadioEvents.TxDone = OnTxDone;
    RadioEvents.TxTimeout = OnTxTimeout;
    RadioEvents.RxDone = OnRxDone;

    Radio.Init( &RadioEvents );
    Radio.SetChannel( RF_FREQUENCY );
    Radio.SetTxConfig( MODEM_LORA, TX_OUTPUT_POWER, 0, LORA_BANDWIDTH,
                                   LORA_SPREADING_FACTOR, LORA_CODINGRATE,
                                   LORA_PREAMBLE_LENGTH, LORA_FIX_LENGTH_PAYLOAD_ON,
                                   true, 0, 0, LORA_IQ_INVERSION_ON, 3000 );

    Radio.SetRxConfig( MODEM_LORA, LORA_BANDWIDTH, LORA_SPREADING_FACTOR,
                                   LORA_CODINGRATE, 0, LORA_PREAMBLE_LENGTH,
                                   LORA_SYMBOL_TIMEOUT, LORA_FIX_LENGTH_PAYLOAD_ON,
                                   0, true, 0, 0, LORA_IQ_INVERSION_ON, true );
    state=TX;
    
  Serial.println("[setup] LoRa initialisation complete");
  Serial.println();
  
  Serial.println("[setup] Set-up Complete");
  Serial.println(""); 

  //  Let everything settle down before continuing
  delay(500);
}

void loop()
{
	switch(state)
	{
		case TX:
			delay(1000);
//			txnumber++;
//		    sprintf(txpacket,"%s","hello");
//		    sprintf(txpacket+strlen(txpacket),"%d",txnumber);
//		    sprintf(txpacket+strlen(txpacket),"%s"," rssi : ");
//		    sprintf(txpacket+strlen(txpacket),"%d",RSSI);
		    RGB_ON(20,0);

//		    Serial.printf("\r\nsending packet \"%s\" , length %d\r\n",txpacket, strlen(txpacket));

//		    Radio.Send( (uint8_t *)txpacket, strlen(txpacket) );
//		    state=LOWPOWER;

        readSensor();
        
        Serial.println("[loop] Send Message");
        sendMessage(windTypePacket, messageCounter);
        Serial.println("[loop] Increment Message Counter");
        messageCounter++;                 // Increment message count
        Serial.println("[loop] Delay...");
        Serial.println(""); 
        RGB_ON(0,0);
        delay(reportInterval);
       
		    break;
		case RX:
			Serial.println("into RX mode");
		    Radio.Rx( 0 );
		    state=LOWPOWER;
		    break;
		case LOWPOWER:
			LowPower_Handler();
		    break;
    default:
        break;
	}
//    Radio.IrqProcess( );
}

void OnTxDone( void )
{
	Serial.print("TX done......");
	RGB_ON(0,0);
//	state=RX;
}

void OnTxTimeout( void )
{
    Radio.Sleep( );
    Serial.print("TX Timeout......");
    state=TX;
}


void OnRxDone( uint8_t *payload, uint16_t size, int16_t rssi, int8_t snr )
{
    RSSI=rssi;
    rxSize=size;
    memcpy(rxpacket, payload, size );
    rxpacket[size]='\0';
    RGB_ON(COLOR_RECEIVED,0);
    Radio.Sleep( );

    Serial.printf("\r\nreceived packet \"%s\" with RSSI %d , length %d\r\n",rxpacket,RSSI,rxSize);
    Serial.println("wait to send next packet");

    state=TX;
}

void readSensor() {
/*
 * IF this node is readin atmospheric conditions from a BME sensor
 * 
  temperature = (int) (10*bme280.getTemperature());
  pressure = (int) (bme280.getPressure() / 91.79F);
  humidity = (int) (bme280.getHumidity());

  Serial.println("");
  Serial.print("[readSensor] Temperature: ");
  Serial.println((float) temperature/10, 1);
  Serial.print("[readSensor] Pressure: ");
  Serial.println(pressure);
  Serial.print("[readSensor] Humidity: ");
  Serial.println(humidity);
*/

/*
 * When we've actually got a sensor to read, this variable should be set to the value
 * read from the sensor
 *
//  inputRainfall = (int) adc0 * LSB;
  inputRainfall = 155; // Dummy value (in 1/10 mm) just so taht there's somethign other than 0 to report

  Serial.print("[readSensor] Rainfall: ");
  Serial.println((float) inputRainfall/10, 1);
*/

/*
 * When we've actually got a sensor to read, these variables should be set to the values
 * read from the sensor(s)
 */
  windDirection = 270;
  windSpeed = 10;
/*
  Serial.print("[readSensor] Wind Direction: ");
  Serial.println(windDirection);
  Serial.print("[readSensor] Wind Speed: ");
  Serial.println(windSpeed);
*/
}

void sendMessage(uint8_t packetType, int counter) {

  outgoingPacket.content.destinationMAC = gatewayMAC;
  outgoingPacket.content.sourceMAC = myMAC;
  outgoingPacket.content.sequenceNumber = (uint16_t) counter % 65536;
  outgoingPacket.content.type = packetType;
  outgoingPacket.content.byteCount = (uint8_t) bytesIn(packetType);

  switch (packetType) {
    case powerTypePacket:
// Nothing here yet
      break;
    case tankTypePacket:
// Nothing here yet
      break;
    case pumpTypePacket:
// Nothing here yet
      break;
    case weatherTypePacket:
// Nothing here yet
      break;
    case atmosphereTypePacket:
    {
      atmospherePayload payload;
      payload.recorded.temperature = temperature;
      payload.recorded.pressure = pressure;
      payload.recorded.humidity = humidity;
      Serial.print("[sendMessage] Sending ");
      Serial.print(myName);
      Serial.print(" atmospheric sensor message - Sequence Number ");
      Serial.println(counter % 65536);
      for (int i = 0; i < outgoingPacket.content.byteCount; i++) {
        outgoingPacket.content.payload[i] = payload.payloadByte[i];
      }
      break;
    }
    case rainfallTypePacket:
// Nothing here yet
      break;
    case windTypePacket:
    {
      windPayload payload;
      payload.recorded.windDirection = windDirection;
      payload.recorded.windSpeed = windSpeed;
      Serial.print("[sendMessage] Sending ");
      Serial.print(myName);
      Serial.print(" wind sensor message - Sequence Number ");
      Serial.println(counter % 65536);
      for (int i = 0; i < outgoingPacket.content.byteCount; i++) {
        outgoingPacket.content.payload[i] = payload.payloadByte[i];
      }
      break;
    }
    case sprinklerTypePacket:
// Nothing here yet
      break;
    default:
      Serial.println("[sendMessage] No payload...");
  }

// Finally, calculate the checksum

 
  outgoingPacket.content.checksum = generateChecksum(outgoingPacket);

  Serial.println("[sendMessage] Outgoing packet: ");
//  digitalWrite(DevKit_Builtin_LED, LOW);  // Turn on the LED to signal sending
/*  
// Check that no one else is transmitting first
  while ((bool) LoRa.available()) {
    Serial.println("[sendMessage] Busy...");
// It would take 212 ms to transmit a [max size] 255 byte packet at 9600 bps
    delay(random(200));
  }
*/  
//  LoRa.beginPacket();             // start packet
  int totalByteCount = LGO_HeaderSize + outgoingPacket.content.byteCount;
  Serial.print("[sendMessage] Byte Count: ");
  Serial.println( totalByteCount );
  for (int i = 0; i < totalByteCount; i++) {
//    LoRa.write(outgoingPacket.packetByte[i]);
    Serial.printf( "%02X", outgoingPacket.packetByte[i]);
    if ((i % 4) == 3) {
      Serial.println("    ");
    } else {
      Serial.print("  ");
    }
  }
  if ((totalByteCount % 4) != 0) Serial.println();
  
  Serial.println("[sendMessage] Finished Packet");
  Radio.Send( (uint8_t *)outgoingPacket.packetByte, totalByteCount );
//  LoRa.endPacket();             // Finish packet and send it
  Serial.println("[sendMessage] Packet Sent");
//  digitalWrite(DevKit_Builtin_LED, HIGH);  // Done
  Serial.println();
}

uint32_t generateChecksum(dataPacket& outgoingPacket) {
  
//  Generate the checksum on the outgoing packet

  CRC32 crc;
  int checksumStart = LGO_HeaderSize - 4;
  int checksumEnd = outgoingPacket.content.byteCount + 4;
  for (int i = checksumStart; i < checksumEnd; i++) {
    crc.update(outgoingPacket.packetByte[i]);
  }
  uint32_t packetChecksum = crc.finalize();
  return packetChecksum;
}

int bytesIn(uint8_t packetType) {
  int result;
  switch (packetType) {
    case powerTypePacket:
      result = powerPayloadBytes;
      break;
    case tankTypePacket:
      result = tankPayloadBytes;
      break;
    case pumpTypePacket:
      result = pumpPayloadBytes;
      break;
    case weatherTypePacket:
      result = weatherPayloadBytes;
      break;
    case atmosphereTypePacket:
      result = atmospherePayloadBytes;
      break;
    case rainfallTypePacket:
      result = rainfallPayloadBytes;
      break;
    case windTypePacket:
      result = windPayloadBytes;
      break;
    case sprinklerTypePacket:
      result = sprinklerPayloadBytes;
      break;
    default:
      result = defaultPayloadBytes;
  }
  return result;
}
