#include "ESP32DevKitPins.h"  // ESP32 Dev Kit Pin Definitions
#include "SPI.h"              // SPI Interface for RFM95
#include "LoRa.h"             // LoRa class & methods
#include "CRC32.h"            // Checksum methods
#include "Adafruit_BME280.h"  // BME280 class & methods
#include "Adafruit_Sensor.h"  // BME280 sensor support methods
#include "Wire.h"             // Need this to define second I2C bus
#include "SSD1306Wire.h"      // OLED Display
#include "SketchFonts.h"      // OLED display fonts
#include "SketchImages.h"     // OLED display images
#include "Ticker.h"           // Ticker class & methods
#include "LangloLoRa.h"       // LoRa configuration parameters
#include "LangloPackets.h"    // Packet structures
#include "nodeA2config.h"     // Node-specific configuration parameters

// ESP32 Dev_Kit to RFM95 (via SPI)
// 3V3    - VCC
// GND    - GND
// GPIO18 - SCK
// GPIO19 - MISO
// GPIO23 - MOSI
// GPIO5  - NSS
// GPIO14 - RST
// GPIO26 - DIO0
// GPIO33 - DIO1
// GPIO32 - DIO2

#define NSS  DevKit_GPIO5
#define SCK  DevKit_GPIO18
#define MISO DevKit_GPIO19
#define MOSI DevKit_GPIO23

#define RST DevKit_GPIO14
#define DI0 DevKit_GPIO26

// Packet structures
dataPacket outgoingPacket = {0};
weatherPayload packetPayload = {0};

// ESP32 Dev_Kit to OLED (via I2C)
// GND    - GND
// 3V3    - VCC
// GPIO22 - SCL
// GPIO21 - SDA

// OLED Display Settings
const int I2C_Display_Address = 0x3c;
const int OLED_I2C_SDA = DevKit_GPIO16;
const int OLED_I2C_SCL = DevKit_GPIO17;

// Define OLED Display
SSD1306Wire     display(I2C_Display_Address, OLED_I2C_SDA, OLED_I2C_SCL);

// ESP32 Dev_Kit to BME280 (via I2C)
// GND    - GND
// 3V3    - VCC
// GPIO21 - SDA
// GPIO22 - SCL

// BME280 Settings
const int I2C_BME_Address = 0x77;   // Not actually necessary — set within Adafruit library
const int BME_I2C_SDA = DevKit_GPIO21;
const int BME_I2C_SCL = DevKit_GPIO22;

// Define BME280 sensor
TwoWire bmeI2C = TwoWire(1);        // Define a second I2C bus to avoid conflict witht he OLED display
Adafruit_BME280 bme280;             // Defaults (I2C_BME_Address = 0x77) set within Adafruit libraries

uint16_t humidity = 0;
uint16_t pressure = 0;
uint16_t temperature = 0;

const float LSB = 0.02761647; // per manual calibration (when using int in 1/10mm)
uint16_t inputRainfall = 0;
int adc0 = 0;

uint16_t windDirection = 0;
uint8_t windSpeed = 0;

const int reportInterval = 60000;     // Report values every 60000 milliseconds (60 seconds)

// Define the update timer
Ticker dataTicker;

int messageCounter = 0;

void setup() {
  pinMode(DevKit_Builtin_LED,OUTPUT);     // Will use LED to signal packet transmission
  digitalWrite(DevKit_Builtin_LED, HIGH);  // Turn it off to begin with (although, I don't know why HIGH is off...
  
  Serial.begin(115200);
  while (!Serial);        // If just the the basic function, must connect to a computer
    delay(500);
  Serial.print("[setup] ");
  Serial.print(myName);
  Serial.println(" I2C BME280/LoRa Sensor");
  Serial.println();

// Initialise the OLED display
  display.init();
//  display.flipScreenVertically();
  display.setFont(ArialMT_Plain_10);
  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(5,5,"BME280/LoRa Sensor");
  display.display();
  
//  Initialise the RFM95 radio
  SPI.begin(SCK,MISO,MOSI,NSS);
  LoRa.setPins(NSS,RST,DI0);
  Serial.println("[setup] Activate LoRa Sender");
  if (!LoRa.begin(BAND, PABOOST)) {
    Serial.println("[setup] Starting LoRa failed!");
    while (true);
  }

  Serial.print("[setup] LoRa Frequency: ");
  Serial.println(BAND);
  
  LoRa.setSpreadingFactor(SpreadingFactor);
  LoRa.setSignalBandwidth(SignalBandwidth);
  LoRa.setPreambleLength(PreambleLength);
  LoRa.setCodingRate4(CodingRateDenominator);
  LoRa.setSyncWord(SyncWord);
  LoRa.disableCrc();
  
  Serial.println("[setup] LoRa initialisation complete");
  Serial.println();
  display.drawString(5,20,"LoRa initialisation OK");
  display.display();

// Initialise the BME280 sensor
  Serial.println("[setup] Initialise BME280 sensor");
  bmeI2C.begin(BME_I2C_SDA, BME_I2C_SCL, 400000);
  if (bme280.begin(I2C_BME_Address, &bmeI2C)) {
    Serial.println("[setup] BME280 sensor initialisation complete");
  } else {
    Serial.println("[setup] Could not find a valid BME280 sensor, check wiring!");
    while (true);
  }
  Serial.println();

  Serial.println("[setup] Set-up Complete");
  Serial.println(""); 

  //  Let everything settle down before continuing
  delay(500);
}

void loop() {
  readSensor();
  Serial.println("[loop] Send Message");
  sendMessage(weatherTypePacket, messageCounter);
  Serial.println("[loop] Update Display");
  updateDisplay(&display, messageCounter);  
  Serial.println("[loop] Increment Message Counter");
  messageCounter++;                 // Increment message count
  Serial.println("[loop] Delay...");
  Serial.println(""); 
  delay(reportInterval);
}

void readSensor() {
  temperature = (int) (10*bme280.readTemperature());
  pressure = (int) (bme280.readPressure() / 91.79F);
  humidity = (int) (bme280.readHumidity());

  Serial.println("");
  Serial.print("[readSensor] Temperature: ");
  Serial.println((float) temperature/10, 1);
  Serial.print("[readSensor] Pressure: ");
  Serial.println(pressure);
  Serial.print("[readSensor] Humidity: ");
  Serial.println(humidity);

/*
 * When we've actually got a sensor to read, this variable should be set to the value
 * read from the sensor
 */
//  inputRainfall = (int) adc0 * LSB;
  inputRainfall = 155; // Dummy value (in 1/10 mm) just so taht there's somethign other than 0 to report

  Serial.print("[readSensor] Rainfall: ");
  Serial.println((float) inputRainfall/10, 1);

/*
 * When we've actually got a sensor to read, these variables should be set to the values
 * read from the sensor(s)
 */
  windDirection = 270;
  windSpeed = 10;

  Serial.print("[readSensor] Wind Direction: ");
  Serial.println(windDirection);
  Serial.print("[readSensor] Wind Speed: ");
  Serial.println(windSpeed);
}

void sendMessage(uint8_t packetType, int counter) {

// If the host is little endian, we need to reverse the byte order before sending

  outgoingPacket.content.destinationMAC = gatewayMAC;
  outgoingPacket.content.sourceMAC = myMAC;
  outgoingPacket.content.sequenceNumber = (uint16_t) counter % 65536;
  outgoingPacket.content.type = packetType;
  outgoingPacket.content.byteCount = (uint8_t) bytesIn(packetType);

  switch (packetType) {
    case powerTypePacket:
// Nothing here yet
      break;
    case tankTypePacket:
// Nothing here yet
      break;
    case pumpTypePacket:
// Nothing here yet
      break;
    case weatherTypePacket:
    {
      weatherPayload payload;
      payload.recorded.temperature = temperature;
      payload.recorded.pressure = pressure;
      payload.recorded.humidity = humidity;

      payload.recorded.rainfall = inputRainfall;

      payload.recorded.windDirection = windDirection;
      payload.recorded.windSpeed = windSpeed;
      
      Serial.print("[sendMessage] Sending ");
      Serial.print(myName);
      Serial.print(" weather sensor message - Sequence Number ");
      Serial.println(counter % 65536);
      for (int i = 0; i < outgoingPacket.content.byteCount; i++) {
        outgoingPacket.content.payload[i] = payload.payloadByte[i];
      }
      break;
    }
    case atmosphereTypePacket:
    {
      atmospherePayload payload;
      payload.recorded.temperature = temperature;
      payload.recorded.pressure = pressure;
      payload.recorded.humidity = humidity;
      Serial.print("[sendMessage] Sending ");
      Serial.print(myName);
      Serial.print(" atmospheric sensor message - Sequence Number ");
      Serial.println(counter % 65536);
      for (int i = 0; i < outgoingPacket.content.byteCount; i++) {
        outgoingPacket.content.payload[i] = payload.payloadByte[i];
      }
      break;
    }
    case rainfallTypePacket:
// Nothing here yet
      break;
    case windTypePacket:
// Nothing here yet
      break;
    case sprinklerTypePacket:
// Nothing here yet
      break;
    default:
      Serial.println("[sendMessage] No payload...");
  }

// Finally, calculate the checksum

  outgoingPacket.content.checksum = generateChecksum(outgoingPacket);

  Serial.println("[sendMessage] Outgoing packet: ");
  digitalWrite(DevKit_Builtin_LED, LOW);  // Turn on the LED to signal sending
/*  
// Check that no one else is transmitting first
  while ((bool) LoRa.available()) {
    Serial.println("[sendMessage] Busy...");
// It would take 212 ms to transmit a [max size] 255 byte packet at 9600 bps
    delay(random(200));
  }
*/  
  LoRa.beginPacket();             // start packet
  int totalByteCount = LGO_HeaderSize + outgoingPacket.content.byteCount;
  Serial.print("[sendMessage] Byte Count: ");
  Serial.println( totalByteCount );
  for (int i = 0; i < totalByteCount; i++) {
    LoRa.write(outgoingPacket.packetByte[i]);
    Serial.printf( "%02X", outgoingPacket.packetByte[i]);
    if ((i % 4) == 3) {
      Serial.println("    ");
    } else {
      Serial.print("  ");
    }
  }
  if ((totalByteCount % 4) != 0) Serial.println();

  Serial.println("[sendMessage] Finished Packet");
  LoRa.endPacket();             // Finish packet and send it
  Serial.println("[sendMessage] Packet Sent");
  digitalWrite(DevKit_Builtin_LED, HIGH);  // Done
  Serial.println();
}

uint32_t generateChecksum(dataPacket& outgoingPacket) {
  
//  Generate the checksum on the outgoing packet

  CRC32 crc;
  int checksumStart = LGO_HeaderSize - 4;
  int checksumEnd = outgoingPacket.content.byteCount + 4;
  for (int i = checksumStart; i < checksumEnd; i++) {
    crc.update(outgoingPacket.packetByte[i]);
  }
  uint32_t packetChecksum = crc.finalize();
  return packetChecksum;
}

int bytesIn(uint8_t packetType) {
  int result;
  switch (packetType) {
    case powerTypePacket:
      result = powerPayloadBytes;
      break;
    case tankTypePacket:
      result = tankPayloadBytes;
      break;
    case pumpTypePacket:
      result = pumpPayloadBytes;
      break;
    case weatherTypePacket:
      result = weatherPayloadBytes;
      break;
    case atmosphereTypePacket:
      result = atmospherePayloadBytes;
      break;
    case rainfallTypePacket:
      result = rainfallPayloadBytes;
      break;
    case windTypePacket:
      result = windPayloadBytes;
      break;
    case sprinklerTypePacket:
      result = sprinklerPayloadBytes;
      break;
    default:
      result = defaultPayloadBytes;
  }
  return result;
}

void updateDisplay(OLEDDisplay *display, int counter) {
  display->clear();
  display->setTextAlignment(TEXT_ALIGN_CENTER);
  display->setFont(ArialMT_Plain_10);
  display->drawString(64, 0, String(myName) + " Weather");

//  Display the sensor readings
  display->setTextAlignment(TEXT_ALIGN_RIGHT);
  display->drawString(60, 12, "Temperature");
  display->drawString(60, 24, "Pressure");
  display->drawString(60, 36, "Humidity");
  display->setTextAlignment(TEXT_ALIGN_LEFT);
  display->drawString(64, 12, String((float) temperature/10, 1) + " °C");
  display->drawString(64, 24, String(pressure) + " hPa");
  display->drawString(64, 36, String(humidity) + " %");

//  Just display the send message discretely
  display->drawHorizontalLine(0, 50, 128);
  display->drawString(0, 52, "Sending packet");
  display->setTextAlignment(TEXT_ALIGN_RIGHT);
  display->drawString(128, 52, String(counter));

  display->display();
}
