/* 
 *  Arduino Pro Mini/RFM95W with BME280E sensor on Wijnand PCB
 *  Site configuration details in <LangloLoRa.h> (arduino/libraries)
 *  Individual node configurations maaintained in <nodeXconfig.h> (arduino/libraries)
 *  (see also <LangloNodes.h> (arduino/libraries)
 *  Packet structures in <LangloPackets.h> (arduino/libraries)
 *  
 *  Stage 1: Just get the comms working — Done
 *  Stage 2: Configure battery (LiFePO4) – Done
 *  Stage 3: Improve battery life
 *              Include battery monitor - Done
 *              Implement LowPower library - Done
 *  Stage 4: Remove MCU LED - Done
 *           Remove MCU rectifier - Done
 *  Stage 5: More energy saving...?
 */

#include "ProMiniPins.h"	  // Pro Mini Pin Definitions
#include "LowPower.h"         // ? Low Popwer library
#include "SPI.h"              // SPI Interface for RFM95
#include "LoRa.h"             // LoRa class & methods
#include "CRC32.h"            // Checksum methods
#include "Wire.h"             // Need this to define second I2C bus (Only one bus on a Pro Mini?)
#include "Adafruit_BME280.h"  // BME280 class & methods
#include "Adafruit_Sensor.h"  // BME280 sensor support methods
#include "LangloLoRa.h"       // LoRa configuration parameters
#include "LangloPackets.h"    // Packet structures
#include "nodeE1config.h"     // Node-specific configuration parameters
                              //    E1 = Breadboard
                              //    E2 = Not assigned
                              //    E3 - Porch
                              //    E4 - Shed

// Pro Mini to RFM95 (via SPI)
// VCC       - VCC
// GND       - GND
// SCK       - SCK
// MISO      - MISO
// MOSI      - MOSI
// ProMini_6 - NSS
// ProMini_5 - RST
// ProMini_2 - DIO0
// ProMini_3 - DIO1
// ProMini_4 - DIO2

#define NSS  ProMini_6
#define SCK  ProMini_SCK
#define MISO ProMini_MISO
#define MOSI ProMini_MOSI

#define RST ProMini_5
#define DI0 ProMini_2

// Packet structures
dataPacket outgoingPacket = {0};
atmospherePayload packetPayload = {0};

// ProMini to I2C device - OLED & BME280
// GND - GND
// VCC - VCC
// SDA - SDA
// SCL - SCL

#define SDA = ProMini_SDA;
#define SCL = ProMini_SCL;

// ProMini to BME280 (via I2C)
// GND - GND
// VCC - VCC
// SDA - SDA
// SCL- SCL

// BME280 Settings
#define SEALEVELPRESSURE_HPA (1013.25)
const int I2C_BME_Address = 0x76;

// Define BME280 sensor
// TwoWire bmeI2C = TwoWire(1);        // Define a second I2C bus to avoid conflict with the OLED display
Adafruit_BME280 bme280;             // Defaults (I2C_BME_Address = 0x77) set within Adafruit libraries

uint16_t humidity = 0;
uint16_t pressure = 0;
int temperature = 0;

const int reportInterval = 60000;     // Report values every 60000 milliseconds (60 seconds)
const int sleepCycles = 8;   // Can only sleep for 8 seconds at a time

// Power supply details
int vcc;

int messageCounter = 0;

void setup() {
  pinMode(ProMini_Builtin_LED,OUTPUT);     // Will use LED to signal packet transmission
  digitalWrite(ProMini_Builtin_LED, HIGH);  // Turn it off to begin with (although, I don't know why HIGH is off...
  
  Serial.begin(115200);
  while (!Serial);        // If just the the basic function, must connect to a computer
    delay(500);
  Serial.print("[setup] ");
  Serial.print(myName);
  Serial.println(" I2C BME280/LoRa Sensor");
  Serial.println();

//  Initialise the RFM95 radio
//  SPI.begin(SCK,MISO,MOSI,NSS);
  LoRa.setPins(NSS, RST, DI0);
  Serial.println("[setup] Activate LoRa Sender");
//  if (!LoRa.begin(BAND, PABOOST)) {
  if (!LoRa.begin(BAND)) {
    Serial.println("[setup] Starting LoRa failed!");
    while (true);
  }

  Serial.print("[setup] LoRa Frequency: ");
  Serial.println(BAND);
  
  LoRa.setSpreadingFactor(SpreadingFactor);
  LoRa.setSignalBandwidth(SignalBandwidth);
  LoRa.setPreambleLength(PreambleLength);
  LoRa.setCodingRate4(CodingRateDenominator);
  LoRa.setSyncWord(SyncWord);
  LoRa.disableCrc();
  
  Serial.println("[setup] LoRa initialisation complete");
  Serial.println();

// Initialise the BME280 sensor
  Serial.println("[setup] Initialise BME280 sensor");
//  bmeI2C.begin(BME_I2C_SDA, BME_I2C_SCL, 400000);
//  if (bme280.begin(I2C_BME_Address, &bmeI2C)) {
  if (bme280.begin(I2C_BME_Address)) {
    Serial.println("[setup] BME280 sensor initialisation complete");
  } else {
    Serial.println("[setup] Could not find a valid BME280 sensor, check wiring!");
    while (true);
  }
  Serial.println();

  Serial.println("[setup] Set-up Complete");
  Serial.println(""); 

  //  Let everything settle down before continuing
  delay(500);
}

void loop() {
  Serial.println("[loop] Send Messages");
  Serial.println("");
  vcc = (int) (1000*readVcc());
  sendMessage(voltageTypePacket, messageCounter);
  // Have a snooze while the gateway porocesses that one
  LowPower.powerDown(SLEEP_1S, ADC_OFF, BOD_OFF);
  readSensor();
  Serial.println("[loop] Send Message");
  sendMessage(atmosphereTypePacket, messageCounter);
  Serial.println("[loop] Increment Message Counter");
  messageCounter++;                 // Increment message count

//  Serial.println("[loop] Delay...");
//  Serial.println("");
//  It seems that 8s is the longest sleep, so need to loop to get to 5 min
//  delay(60000);

  for ( int i = 0; i < sleepCycles; ++i ) {
//    Serial.print("[loop] Sleep cycle ");
//    Serial.println(i);
    LowPower.powerDown(SLEEP_8S, ADC_OFF, BOD_OFF);
  }
}

float readVcc(void) {
  float result = 0;

  ADMUX = _BV(REFS0) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);
  delay(2);                         // Wait for Vref to settle

  ADCSRA |= _BV(ADSC);              // Start conversion
  while (bit_is_set(ADCSRA,ADSC));  // Wait conversion is ready
 
  uint8_t low  = ADCL; // Read ADCL first - it then locks ADCH  
  uint8_t high = ADCH; // unlocks both
  result = (float)((high<<8) | low);
 
  result = 1125300L/result;  // Calculate battery voltage in mV
  Serial.print("[readVcc] Voltage: ");
  Serial.print(result/1000);
  Serial.println(" V");
  return (result/1000);      // Return battery voltage in V
};

void readSensor() {
  temperature = (int) (10*bme280.readTemperature());
  pressure = (int) (bme280.readPressure() / 91.79F);
  humidity = (int) bme280.readHumidity();

  Serial.println("[updateData] Sensor read...");
  Serial.print("[updateData] Temperature: ");
  Serial.print((float) temperature/10, 1);
  Serial.println(" °C");
  Serial.print("[updateData] Pressure: ");
  Serial.print(pressure);
  Serial.println(" hPa");
  Serial.print("[updateData] Humidity: ");
  Serial.print(humidity);
  Serial.println(" %");
  Serial.println("");
}

void sendMessage(uint8_t packetType, int counter) {

// If the host is little endian, we need to reverse the byte order before sending

  outgoingPacket.content.destinationMAC = gatewayMAC;
  outgoingPacket.content.sourceMAC = myMAC;
  outgoingPacket.content.sequenceNumber = (uint16_t) counter % 65536;
  outgoingPacket.content.type = packetType;
  outgoingPacket.content.byteCount = (uint8_t) bytesIn(packetType);

  switch (packetType) {
    case powerTypePacket:
// Nothing here yet
      break;
    case voltageTypePacket:
    {
      voltagePayload payload;
      payload.source.voltage = vcc;
      Serial.print("[sendMessage] Sending ");
      Serial.print(myName);
      Serial.print(" voltage message - Sequence Number ");
      Serial.println(counter % 65536);
      for (int i = 0; i < outgoingPacket.content.byteCount; i++) {
        outgoingPacket.content.payload[i] = payload.payloadByte[i];
      }
      break;
    }
    case tankTypePacket:
// Nothing here yet
      break;
    case pumpTypePacket:
// Nothing here yet
      break;
    case weatherTypePacket:
// Nothing here yet
      break;
    case atmosphereTypePacket:
    {
      atmospherePayload payload;
      payload.recorded.temperature = temperature;
      payload.recorded.pressure = pressure;
      payload.recorded.humidity = humidity;
      Serial.print("[sendMessage] Sending ");
      Serial.print(myName);
      Serial.print(" atmospheric sensor message - Sequence Number ");
      Serial.println(counter % 65536);
      for (int i = 0; i < outgoingPacket.content.byteCount; i++) {
        outgoingPacket.content.payload[i] = payload.payloadByte[i];
      }
     break;
    }
    case rainfallTypePacket:
// Nothing here yet
      break;
    case windTypePacket:
// Nothing here yet
      break;
    case sprinklerTypePacket:
// Nothing here yet
      break;
    default:
      Serial.println("[sendMessage] No payload...");
  }

// Finally, calculate the checksum

  outgoingPacket.content.checksum = generateChecksum(outgoingPacket);

  Serial.println("[sendMessage] Outgoing packet: ");
  digitalWrite(ProMini_Builtin_LED, LOW);  // Turn on the LED to signal sending
/*  
// Check that no one else is transmitting first
  while ((bool) LoRa.available()) {
    Serial.println("[sendMessage] Busy...");
// It would take 212 ms to transmit a [max size] 255 byte packet at 9600 bps
    delay(random(200));
  }
*/  
  LoRa.beginPacket();             // start packet
  int totalByteCount = LGO_HeaderSize + outgoingPacket.content.byteCount;
  Serial.print("[sendMessage] Byte Count: ");
  Serial.println( totalByteCount );
  for (int i = 0; i < totalByteCount; i++) {
    LoRa.write(outgoingPacket.packetByte[i]);
    if (outgoingPacket.packetByte[i] < 0x10) { Serial.print("0"); } 
    Serial.print( outgoingPacket.packetByte[i], HEX);
    if ((i % 4) == 3) {
      Serial.println("    ");
    } else {
      Serial.print("  ");
    }
  }
  if (totalByteCount != 0) Serial.println("");

//  Serial.println("[sendMessage] Finished Packet");
  LoRa.endPacket();             // Finish packet and send it
//  Serial.println("[sendMessage] Packet Sent");
  digitalWrite(ProMini_Builtin_LED, HIGH);  // Done
//  Serial.println();
}

uint32_t generateChecksum(dataPacket& outgoingPacket) {
  
//  Generate the checksum on the outgoing packet

  CRC32 crc;
  int checksumStart = LGO_HeaderSize - 4;
  int checksumEnd = outgoingPacket.content.byteCount + 4;
  for (int i = checksumStart; i < checksumEnd; i++) {
    crc.update(outgoingPacket.packetByte[i]);
  }
  uint32_t packetChecksum = crc.finalize();
  return packetChecksum;
}

int bytesIn(uint8_t packetType) {
  int result;
  switch (packetType) {
    case powerTypePacket:
      result = powerPayloadBytes;
      break;
    case voltageTypePacket:
      result = voltagePayloadBytes;
      break;
    case tankTypePacket:
      result = tankPayloadBytes;
      break;
    case pumpTypePacket:
      result = pumpPayloadBytes;
      break;
    case weatherTypePacket:
      result = weatherPayloadBytes;
      break;
    case atmosphereTypePacket:
      result = atmospherePayloadBytes;
      break;
    case rainfallTypePacket:
      result = rainfallPayloadBytes;
      break;
    case windTypePacket:
      result = windPayloadBytes;
      break;
    case sprinklerTypePacket:
      result = sprinklerPayloadBytes;
      break;
    default:
      result = defaultPayloadBytes;
  }
  return result;
}
