#include "LoRa.h"
#include "Arduino.h"
#include "CRC32.h"            // Checksum methods
#include "LangloLoRa.h"       // LoRa configuration parameters
#include "LangloPackets.h"    // Packet structures
#include "HeltecWiFiLoRaV1Pins.h" // Pin identifiers

// Voltage conversion factor
const float LSB = 0.002761647; // per manual calibration
// We sample the voltage two times per sec, then report a rolling ten-sample average
const int inputVoltageDimension = 10;
int adc0 = 0;
int valueIndex = 0;
float inputVoltage[inputVoltageDimension] = {0.0};
String averageVoltage = "";

// MAC Address - Hard coded at this stage only to simplify system development
const uint32_t  myMAC = 0xA16EF75C;
const uint32_t  gatewayMAC = 0xA16EF174;

// WIFI_LoRa_32 ports

// GPIO5 — SX1278’s SCK
// GPIO19 — SX1278’s MISO
// GPIO27 — SX1278’s MOSI
// GPIO18 — SX1278’s CS
// GPIO14 — SX1278’s RESET
// GPIO26 — SX1278’s IRQ(Interrupt Request)

int counter = 0;
dataPacket outgoingPacket;
powerPayload packetPayload;

void setup() {
  LoRa.setPins(CS,RST,IRQ);
  if (!LoRa.begin(BAND, PABOOST)) {
    while (true);
  }
  LoRa.setSpreadingFactor(SpreadingFactor);
  LoRa.setSignalBandwidth(SignalBandwidth);
  LoRa.setPreambleLength(PreambleLength);
  LoRa.setCodingRate4(CodingRateDenominator);
  LoRa.setSyncWord(SyncWord);
  LoRa.disableCrc();

  // The only value currently measured is the battery voltage—just zero the others
  packetPayload.battery.current = 0;
  packetPayload.panel.voltage = 0;
  packetPayload.panel.current = 0;
  packetPayload.load.voltage = 0;
  packetPayload.load.current = 0;
  
  delay(2000);
}

void loop() {

// Read the voltage from ADC1_0 (Pin 36)
  adc0 = analogRead(36);

// Since the number fluctates so much, we calcualte a rolling average and just report that
  if (valueIndex < 0) {
    valueIndex = 0;
  } else if (valueIndex > (inputVoltageDimension - 1)) {
    valueIndex = 0;
  }
  inputVoltage[valueIndex] = adc0 * LSB;
//  Serial.print("  ADC1_0: "); Serial.print(adc0);
//  Serial.print("  Voltage: "); Serial.println(inputVoltage[valueIndex]);
  valueIndex++;
  float averageVoltage = rollingAverage(inputVoltage,inputVoltageDimension);
  String averageVoltageString = String(averageVoltage, 2);

//  Serial.print("[Loop] Sending packet: ");
//  Serial.println(counter);
  
  packetPayload.battery.voltage = (int) (1000 * averageVoltage);

  sendMessage(counter, powerTypePacket, packetPayload.payloadByte);
  counter++;                 // Increment message count
 
 delay(60000);
}

void sendMessage(int counter, uint8_t packetType, uint8_t payload[]) {

  outgoingPacket.content.destinationMAC = gatewayMAC;
  outgoingPacket.content.sourceMAC = myMAC;
  outgoingPacket.content.sequenceNumber = (uint16_t) counter % 65536;
  outgoingPacket.content.type = packetType;
  outgoingPacket.content.byteCount = (uint8_t) bytesIn(packetType);
  for (int i = 0; i < outgoingPacket.content.byteCount; i++) {
    outgoingPacket.content.payload[i] = payload[i];
  }

// Finally, calculate the checksum

  uint32_t outgoingChecksum = generateChecksum(outgoingPacket);
  outgoingPacket.content.checksum = outgoingChecksum;
  
  LoRa.beginPacket();             // start packet
  int totalByteCount = LGO_HeaderSize + outgoingPacket.content.byteCount;
  for (int i = 0; i < totalByteCount; i++) {
    LoRa.write(outgoingPacket.packetByte[i]);
  }
  LoRa.endPacket();             // Finish packet and send it
}

uint32_t generateChecksum(dataPacket& outgoingPacket) {
  
//  Generate the checksum on the outgoing packet

  CRC32 crc;
  int checksumStart = LGO_HeaderSize - 4;
  int checksumEnd = outgoingPacket.content.byteCount + 4;
  for (int i = checksumStart; i < checksumEnd; i++) {
    crc.update(outgoingPacket.packetByte[i]);
  }
  uint32_t packetChecksum = crc.finalize();
  return packetChecksum;
}

int bytesIn(uint8_t packetType) {
  int result;
  switch (packetType) {
    case powerTypePacket:
      result = powerPayloadBytes;
      break;
    case tankTypePacket:
      result = tankPayloadBytes;
      break;
    case pumpTypePacket:
      result = pumpPayloadBytes;
      break;
    case weatherTypePacket:
      result = weatherPayloadBytes;
      break;
    case sprinklerTypePacket:
      result = sprinklerPayloadBytes;
      break;
    default:
      result = defaultPayloadBytes;
  }
  return result;
}

float rollingAverage( float value[], int valueDimension ) {
  float valueSum = 0.0;
  for (int i = 0; i < valueDimension; i++) {
    valueSum = valueSum + value[i];
  }
  return valueSum/valueDimension;
}
